/*
 * bin2abc.c
 *
 * Convert a binary file either in binary or ith format to an ABC80/800
 * .bac or .abs file.
 */

#include "abcdisk.h"
#include "bin2abc.h"

/* Aren't modern computers nice? */
static uint8_t data[MAX_ADDR];
static bool isset[MAX_ADDR];

FILE *open_file(const char *filename, const char *mode)
{
    FILE *f;

    if (is_stdio(filename)) {
        f = (mode[0] == 'w') ? stdout : stdin;
    } else {
        f = fopen(filename, mode);
        if (!f) {
            fprintf(stderr, "%s: %s: %s\n", program_name, filename,
                    strerror(errno));
            return NULL;
        }
    }

    clearerr(f);
    return f;
}

int close_file(FILE * f)
{
    int err = ferror(f);

    if (f != stdin && f != stdout && f != stderr)
        err |= fclose(f);

    return err;
}

size_t read_file(void *buf, size_t buflen, const char *filename)
{
    FILE *f;
    size_t n;

    f = open_file(filename, "rb");
    if (!f)
        return 0;

    n = fread(buf, 1, buflen, f);

    close_file(f);
    return n;
}

static int read_binary(const char *filename, uint16_t offset,
                       uint8_t * buf, bool * isset)
{
    size_t i, n;

    n = read_file(buf + offset, MAX_ADDR - offset, filename);
    if (!n)
        return -1;

    for (i = offset; i < offset + n; i++)
        isset[i] = true;

    return offset + n;
}

const char *program_name;

static no_return usage(void)
{
    fprintf(stderr, "abcdisk %s\n", VERSION);
    fprintf(stderr, "Type %s --help for usage instructions\n", program_name);
    exit(1);
}

/* Helper functions that error out on a missing argument */
static char *short_arg(char opt, char *arg)
{
    if (!arg) {
        fprintf(stderr, "%s: the -%c option requires an argument\n",
                program_name, opt);
        usage();
    }
    return arg;
}

static char *long_arg(const char *opt, char *arg)
{
    if (!arg) {
        fprintf(stderr, "%s: the --%s option requires an argument\n",
                program_name, opt);
        usage();
    }
    return arg;
}

#define SHORT_ARG()	short_arg(optchr, *option++)
#define LONG_ARG()	long_arg(optstr,  *option++)

static unsigned long intarg(const char *arg)
{
    char *ep;
    unsigned long v;

    v = strtoul(arg, &ep, 0);
    if (!ep[0])
        return v;

    fprintf(stderr, "%s: invalid integer argument: %s\n", program_name, arg);
    usage();
}

static no_return help(void)
{
    printf("abcdisk %s\n", VERSION);
    printf("Usage: %s [options...] inputfiles...\n"
           "Options:\n"
           "  -1, --bac80       generate a .bac file for ABC80\n"
           "  -2, --bac800      generate a .bac file for ABC800\n"
           "  -a, --abs         generate an .abs file\n"
           "  -o, --output ...  send output to a file instead of stdout\n"
           "  -r, --reloc ...   read a relocation file (.reloc from z80asm)\n"
           "  -b, --binary      input files are binary\n"
           "  -i, --ihex        input files are in Intel hex format\n"
           "  -d, --driver      generate .bac files suitable for device drivers\n"
           "  -D, --app         generate .bac files suitable for applications\n"
           "  -s, --origin #    specify the load address for a binary file\n"
           "  -e, --entry #     specify the entry point of the program\n"
           "  -z, --bss #       pad the output with # bytes of zero\n"
           "  -h, --help        show this help text\n"
	   "      --version     print the program version\n"
	   , program_name);
    exit(0);
}

enum output {
    OUT_BAC80,
    OUT_BAC800,
    OUT_ABS
};

int main(int argc, char *argv[])
{
    bool ihex = false;
    char **option, *optstr, optchr;
    enum output output = OUT_BAC80;
    int entrypt = -1;
    int bss = 0;
    bool driver = false;
    const char *outputfile = NULL;
    FILE *out;
    int firstaddr, i;
    int lastaddr = 0;
    int err = 0;

    program_name = argv[0];

    (void)argc;

    option = &argv[1];
    while ((optstr = *option++) != NULL) {
        if (optstr[0] == '-') {
            optstr++;
            optchr = *optstr++;

            if (optchr == '-') {
                if (!strcmp(optstr, "abc80") || !strcmp(optstr, "bac80")) {
                    output = OUT_BAC80;
                } else if (!strcmp(optstr, "abc800") ||
                           !strcmp(optstr, "bac800")) {
                    output = OUT_BAC800;
                } else if (!strcmp(optstr, "abs")) {
                    output = OUT_ABS;
                } else if (!strcmp(optstr, "reloc")) {
		    read_reloc_file(LONG_ARG());
                } else if (!strcmp(optstr, "ihex") ||
                           !strcmp(optstr, "ith") ||
                           !strcmp(optstr, "intel") || !strcmp(optstr, "hex")) {
                    ihex = true;
                } else if (!strcmp(optstr, "binary")) {
                    ihex = false;
                } else if (!strcmp(optstr, "entry")) {
                    entrypt = intarg(LONG_ARG());
                } else if (!strcmp(optstr, "driver") ||
                           !strcmp(optstr, "no-app")) {
                    driver = true;
                } else if (!strcmp(optstr, "app") ||
                           !strcmp(optstr, "no-driver")) {
                    driver = false;
                } else if (!strcmp(optstr, "output")) {
                    outputfile = LONG_ARG();
                } else if (!strcmp(optstr, "start") ||
                           !strcmp(optstr, "origin")) {
                    lastaddr = intarg(LONG_ARG());
                } else if (!strcmp(optstr, "entry") ||
                           !strcmp(optstr, "entrypoint")) {
                    entrypt = intarg(LONG_ARG());
                } else if (!strcmp(optstr, "bss") || !strcmp(optstr, "zero")) {
                    bss += intarg(LONG_ARG());
		} else if (!strcmp(optstr, "version")) {
		    printf("abcdisk %s\n", VERSION);
		    exit(0);
                } else if (!strcmp(optstr, "help")) {
                    help();
                } else {
                    fprintf(stderr, "%s: unknown option --%s\n",
                            program_name, optstr);
                    usage();
                }
            } else {
                /* Short option */
                while (optchr) {
                    switch (optchr) {
                    case '1':  /* For "BASIC I" */
                        output = OUT_BAC80;
                        break;
                    case '2':  /* For "BASIC II" */
                        output = OUT_BAC800;
                        break;
                    case 'a':
                        output = OUT_ABS;
                        break;
                    case 'i':
                        ihex = true;
                        break;
                    case 'b':
                        ihex = false;
                        break;
                    case 'r':
                        read_reloc_file(SHORT_ARG());
                        break;
                    case 'e':
                        entrypt = intarg(SHORT_ARG());
                        break;
                    case 's':
                        lastaddr = intarg(SHORT_ARG());
                        break;
                    case 'o':
                        outputfile = SHORT_ARG();
                        break;
                    case 'd':
                        driver = true;
                        break;
                    case 'D':
                        driver = false;
                        break;
                    case 'z':
                        bss += intarg(SHORT_ARG());
                        break;
                    case 'h':
                        help();
                        break;
                    default:
                        fprintf(stderr, "%s: unknown option -%c\n",
                                program_name, optchr);
                        usage();
                        break;
                    }
                    optchr = *optstr++;
                }
            }
        } else {
            /* Not an option, input file */
            int end;

            if (ihex) {
                end = readihex(optstr, data, isset, &entrypt);
            } else {
                end = read_binary(optstr, lastaddr, data, isset);
                if (entrypt < 0 && end >= 0 && end != lastaddr)
                    entrypt = lastaddr;
            }
            if (end >= 0)
                lastaddr = end;
        }
    }

    /* Find the first and last addresses in the data array */
    firstaddr = -1;
    lastaddr = -1;
    for (i = 0; i < MAX_ADDR; i++) {
        if (isset[i]) {
            lastaddr = i;
            if (firstaddr < 0)
                firstaddr = i;
        }
    }

    if (firstaddr < 0 && !bss) {
        fprintf(stderr, "%s: no input data!\n", program_name);
	usage();
    }

    /* Add a zero-filled section after the last data */
    if (bss) {
        uint8_t *dp = data + lastaddr;
        bool *ip = isset + lastaddr;

        memset(dp, 0, bss);
        while (bss--)
            *ip++ = true;

        lastaddr += bss;
    }

    out = open_file(outputfile, "wb");
    if (!out) {
        fprintf(stderr, "%s: %s: %s\n", program_name, outputfile,
                strerror(errno));
        exit(1);
    }

    switch (output) {
    case OUT_BAC80:
        err = writebac80(out, data + firstaddr, lastaddr - firstaddr + 1,
                         firstaddr, entrypt, driver);
        break;

    case OUT_BAC800:
        err = writebac800(out, data + firstaddr, lastaddr - firstaddr + 1,
                          firstaddr, entrypt, driver);
        break;

    case OUT_ABS:
	err = writeabs(out, data, isset, firstaddr, entrypt);
        break;
    }

    err |= close_file(out);

    if (err && !is_stdio(outputfile)) {
	if (errno)
	    fprintf(stderr, "%s: %s: %s\n", program_name, outputfile,
		    strerror(errno));
        remove(outputfile);
    }

    return !!err;
}
